from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from pyranges.core.names import VALID_BY_TYPES, VALID_JOIN_TYPE, VALID_OVERLAP_TYPE
from pyranges.methods.overlap import _both_idxs

if TYPE_CHECKING:
    from pyranges.range_frame.range_frame import RangeFrame


def _both_dfs(
    df: "RangeFrame",
    df2: "RangeFrame",
    *,
    by: VALID_BY_TYPES = None,
    multiple: VALID_OVERLAP_TYPE = "all",
    contained: bool = False,
    slack: int = 0,
    join_type: VALID_JOIN_TYPE,
    suffix: str,
) -> pd.DataFrame:
    _self_indexes, _other_indexes = _both_idxs(
        df=df,
        df2=df2,
        by=by,
        multiple=multiple,
        contained=contained,
        slack=slack,
    )
    expected_columns = [*df.head(0).join(df2.head(0), how="inner", rsuffix=suffix).columns]
    df2.columns = expected_columns[df.shape[1] :]
    _df = df.take(_self_indexes)  # type: ignore[arg-type]
    _df.index = pd.Index(np.arange(len(_df)))
    _df2 = pd.DataFrame(df2).take(_other_indexes)  # type: ignore[arg-type]
    _df2.index = pd.Index(np.arange(len(_df2)))
    j = _df.join(_df2, how="inner")
    if join_type == "inner":
        j.index = pd.Index(_self_indexes)
        return j

    if join_type == "left":
        missing_rows = _missing_rows_left(_self_indexes, df, j)
        return pd.concat([j, missing_rows])

    if join_type == "right":
        missing_rows = _missing_rows_right(_other_indexes, df2, j)
        return pd.concat([j, missing_rows])

    if join_type == "outer":
        missing_rows_l = _missing_rows_left(_self_indexes, df, j)
        missing_rows_r = _missing_rows_right(_other_indexes, df2, j)
        return pd.concat([j, missing_rows_l, missing_rows_r])

    msg = f"Invalid join type: {join_type}"
    raise ValueError(msg)


def _missing_rows_left(_self_indexes, df, j) -> pd.DataFrame:
    missing_indices_left = df.index.difference(_self_indexes)
    j.index = _self_indexes
    return df.reindex(missing_indices_left)


def _missing_rows_right(_other_indexes, df2, j) -> pd.DataFrame:
    missing_indices_right = df2.index.difference(_other_indexes)
    j.index = _other_indexes
    return df2.reindex(missing_indices_right)



import csv
import logging
from pathlib import Path
from typing import Literal, Protocol

import numpy as np
import pandas as pd
from natsort import natsorted  # type: ignore[import]
from pandas.core.frame import DataFrame

from pyranges.core.names import BIGWIG_SCORE_COL, CHROM_COL, END_COL, PANDAS_COMPRESSION_TYPE, START_COL
from pyranges.core.pyranges_helpers import mypy_ensure_pyranges
from pyranges.core.pyranges_main import PyRanges

GTF_COLUMNS_TO_PYRANGES = {
    "seqname": "Chromosome",
    "source": "Source",
    "feature": "Feature",
    "start": "Start",
    "end": "End",
    "score": "Score",
    "strand": "Strand",
    "frame": "Frame",
}

GFF3_COLUMNS_TO_PYRANGES = GTF_COLUMNS_TO_PYRANGES.copy()
GFF3_COLUMNS_TO_PYRANGES["phase"] = GFF3_COLUMNS_TO_PYRANGES.pop("frame")

_ordered_gtf_columns = [
    "seqname",
    "source",
    "feature",
    "start",
    "end",
    "score",
    "strand",
    "frame",
    "attribute",
]
_ordered_gff3_columns = [
    "seqname",
    "source",
    "feature",
    "start",
    "end",
    "score",
    "strand",
    "phase",
    "attribute",
]


logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def _fill_missing(df: DataFrame, all_columns: list[str]) -> DataFrame:
    columns = list(df.columns)

    if set(columns).intersection(set(all_columns)) == set(all_columns):
        return df[all_columns]
    missing = set(all_columns) - set(columns)
    missing_idx = {all_columns.index(m): m for m in missing}
    not_missing = set(columns).intersection(set(all_columns))
    not_missing_ordered = sorted(not_missing, key=all_columns.index)
    outdf = df[not_missing_ordered]

    for idx, _missing in sorted(missing_idx.items()):
        outdf.insert(idx, _missing, ".")

    return outdf


def _bed(df: DataFrame, *, keep: bool) -> DataFrame:
    bed_columns = ["Chromosome", "Start", "End", "Name", "Score", "Strand"]

    outdf = _fill_missing(df, bed_columns)

    noncanonical = list(set(df.columns) - set(bed_columns))
    noncanonical = [c for c in df.columns if c in noncanonical]

    if keep:
        return pd.concat([outdf, df[noncanonical]], axis=1)
    return outdf


def _to_gff_like(
    gr: PyRanges,
    out_format: Literal["gtf", "gff3"],
    path: Path | None = None,
    compression: PANDAS_COMPRESSION_TYPE = None,
    map_cols: dict | None = None,
) -> str | None:
    df = _pyranges_to_gtf_like(
        gr,
        out_format=out_format,
        map_cols=map_cols,
    )
    return df.to_csv(
        path,
        index=False,
        header=False,
        compression=compression,
        mode="w",
        sep="\t",
        quoting=csv.QUOTE_NONE,
    )


def _to_csv(
    self: PyRanges,
    path: Path | str | None = None,
    sep: str = ",",
    compression: PANDAS_COMPRESSION_TYPE = None,
    *,
    header: bool = True,
) -> str | None:
    gr = self

    if path:
        mode = "w+"
        for _, outdf in natsorted(gr.dfs.items()):
            outdf.to_csv(
                path,
                index=False,
                compression=compression,
                header=header,
                mode=mode,
                sep=sep,
                quoting=csv.QUOTE_NONE,
            )
            mode = "a"
            header = False
        return None
    return "".join(
        [
            outdf.to_csv(index=False, header=header, sep=sep, quoting=csv.QUOTE_NONE)
            for _, outdf in sorted(gr.dfs.items())
        ],
    )


def _to_bed(
    self: PyRanges,
    path: str | None = None,
    compression: PANDAS_COMPRESSION_TYPE = None,
    *,
    keep: bool = True,
) -> str | None:
    df = _bed(self, keep=keep)

    return df.to_csv(
        path,
        index=False,
        header=False,
        compression=compression,
        mode="w+",
        sep="\t",
        quoting=csv.QUOTE_NONE,
    )


def _to_bigwig(
    self: PyRanges,
    path: None,
    chromosome_sizes: PyRanges | pd.DataFrame | dict,
    value_col: str | None = None,
    *,
    divide: bool = False,
    rpm: bool = True,
    dryrun: bool = False,
) -> PyRanges | None:
    try:
        import pyBigWig  # type: ignore[import]
    except ModuleNotFoundError:
        LOGGER.exception(
            "pybigwig must be installed to create bigwigs. Use `conda install -c bioconda pybigwig` or `pip install pybigwig` to install it.",
        )
        import sys

        sys.exit(1)

    if not divide:
        rles = self.to_rle(rpm=rpm, strand=False, value_col=value_col)
        df = rles.to_ranges()
    else:
        df = self.to_rle(rpm=rpm, strand=False, value_col=value_col)
        divide_by = self.to_rle(rpm=rpm, strand=False)
        c = df / divide_by
        new_pyrles = {}
        for k, v in c.items():
            v.values = np.log2(v.values)
            v.defragment()
            new_pyrles[k] = v

        df = c.defragment().to_ranges()
    gr = mypy_ensure_pyranges(df)
    unique_chromosomes = gr.chromosomes

    gr = gr.remove_strand()
    gr = gr.sort_ranges()
    gr = gr.get_with_loc_columns(BIGWIG_SCORE_COL)

    if dryrun:
        return gr

    if not isinstance(chromosome_sizes, dict):
        size_df = chromosome_sizes
        chromosome_sizes = dict(zip(size_df[CHROM_COL], size_df[END_COL], strict=True))

    header = [(c, int(chromosome_sizes[c])) for c in unique_chromosomes]

    bw = pyBigWig.open(path, "w")
    bw.addHeader(header)

    chromosomes = df[CHROM_COL].tolist()
    starts = df[START_COL].tolist()
    ends = df[END_COL].tolist()
    values = df.Score.tolist()

    bw.addEntries(chromosomes, starts, ends=ends, values=values)

    return None


class AttributeFormatter(Protocol):
    def __call__(self, colname: str, col: "pd.Series[str]") -> "pd.Series[str]":
        """Stub to properly annotate forced named args (..., *, ...)."""
        ...


def _pyranges_to_gtf_like(
    df: pd.DataFrame,
    out_format: Literal["gtf", "gff3"],
    map_cols: dict | None = None,
) -> pd.DataFrame:
    attribute_formatter: AttributeFormatter

    if out_format == "gtf":
        all_columns = _ordered_gtf_columns[:-1]
        # first: gff column to pyranges column
        rename_columns = GFF3_COLUMNS_TO_PYRANGES.copy()
        attribute_formatter = gtf_formatter
    elif out_format == "gff3":
        all_columns = _ordered_gff3_columns[:-1]
        # first: gff column to pyranges column
        rename_columns = GTF_COLUMNS_TO_PYRANGES.copy()
        attribute_formatter = gff3_formatter
    else:
        msg = f"Invalid output format: {out_format}. Must be one of 'gtf' or 'gff3'."
        raise ValueError(msg)

    map_cols = map_cols or {}
    valid_keys = set(rename_columns) | {"attribute"}
    invalid_map_cols = set(map_cols) - valid_keys
    if invalid_map_cols:
        msg = f"Invalid column mapping: {invalid_map_cols}. Must be one of {set(rename_columns) | {'attribute'}}."
        raise ValueError(msg)

    rename_columns.update(map_cols)
    # from here on, rename_columns is: pyranges column to gff column
    rename_columns = {v: k for k, v in rename_columns.items()}

    df = df.rename(columns=rename_columns)
    df.loc[:, "start"] = df.start + 1

    columns = list(df.columns)

    # filling missing columns with "."
    outdf = _fill_missing(df, all_columns)

    if "attribute" not in map_cols:
        _rest = set(df.columns) - set(all_columns)
        rest = sorted(_rest, key=columns.index)
        rest_df = df[rest].copy()
        # putting all remaining columns into the attribute column
        for colname in rest_df.columns:
            col = pd.Series(rest_df[colname])
            isnull = col.isna()
            new_val = attribute_formatter(colname, col)  # type: ignore[call-arg]

            # not working to convert cat to str:  rest_df.loc[:, colname] = rest_df[colname].astype(str)
            # so doing this instead:
            rest_df[colname] = rest_df[colname].astype(str)
            rest_df.loc[~isnull, colname] = new_val
            rest_df.loc[isnull, colname] = ""

        attribute = merge_attributes(rest_df, out_format)
        outdf.insert(outdf.shape[1], column="attribute", value=attribute)
    else:
        outdf.insert(outdf.shape[1], column="attribute", value=df["attribute"].copy())

    return outdf


def gtf_formatter(colname: str, col: "pd.Series[str]") -> "pd.Series[str]":
    """Format a column as a GTF attribute column."""
    attribute_template_string = f'{colname} "{{col}}"; '
    return col.apply(lambda x: attribute_template_string.format(col=x))


def gff3_formatter(colname: str, col: "pd.Series[str]") -> "pd.Series[str]":
    """Format a column as a GFF3 attribute column."""
    attribute_template_string = f"{colname}={{col}};"
    return col.apply(lambda x: attribute_template_string.format(col=x))


def merge_attributes(attributes: pd.DataFrame, out_format: Literal["gtf", "gff3"]) -> "pd.Series[str]":
    """Merge attributes into a single column.

    The final column in gtf/gff is not separated by tabs, but by other separators.
    """
    regex_to_remove = " $" if out_format == "gtf" else ";$"
    return attributes.apply(lambda r: "".join([v for v in r if v]), axis=1).str.replace(regex_to_remove, "", regex=True)



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import ruranges

from pyranges.core.names import (
    END_COL,
    FORWARD_STRAND,
    START_COL,
    STRAND_COL,
)
from pyranges.core.pyranges_helpers import factorize

if TYPE_CHECKING:
    from pyranges import PyRanges


def _subseq(
    df: "PyRanges",
    *,
    by: list[str],
    start: int = 0,
    end: int | None = None,
    force_plus_strand: bool = False,
) -> pd.DataFrame:
    if df.empty:
        return df

    chrs = factorize(df, by)

    outidx, outstarts, outends = ruranges.subsequence_numpy(  # type: ignore[attr-defined]
        chrs,
        df[START_COL].to_numpy(),
        df[END_COL].to_numpy(),
        (df[STRAND_COL] == FORWARD_STRAND).to_numpy()
        if (STRAND_COL in df and not force_plus_strand)
        else np.ones(len(df), dtype=bool),
        start=start,
        end=end,
        force_plus_strand=force_plus_strand,
    )

    outdf = df.take(outidx)
    outdf.loc[:, START_COL] = outstarts
    outdf.loc[:, END_COL] = outends

    return outdf



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from docutils import nodes, utils

# sys.path.insert(0, os.path.abspath('.'))

# sys.path.insert(0, os.path.abspath('..'))  # Adjust this as necessary
#
# sys.path.insert(0, os.path.abspath('../pyranges'))  # Adjust this as necessary
#
# sys.path.insert(0, os.path.abspath('.'))  # Adjust this as necessary

# -- Project information -----------------------------------------------------

project = "pyranges"
copyright = "2024, Endre Bakken Stovner, Marco Mariotti"
author = "Endre Bakken Stovner, Marco Mariotti"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.


extensions = [
    "sphinx.ext.napoleon",
    # "sphinxcontrib.napoleon",
    # "autoapi.extension",
    "sphinx.ext.autodoc",
    # "autoapi.extension",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    # "sphinx.ext.doctest"
]

autosummary_generate = True  # Enable summary table generation


autodoc_default_options = {
    "members": True,
    "imported-members": True,
    # other options...
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_generated_hidden*"]

master_doc = "index"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = "alabaster"
html_theme = "sphinx_rtd_theme"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def monospaced_link(name, rawtext, text, lineno, inliner, options=None, content=None):
    if content is None:
        content = []
    if options is None:
        options = {}
    url = text.split(" ")[-1].strip("<>")
    clickable_text = " ".join(text.split(" ")[:-1])
    # Create a reference node, which is the docutils node for hyperlinks
    utils.unescape(text)

    node = nodes.reference(rawtext, clickable_text, refuri=url, **options)

    # Add a special class to this node
    node["classes"].append("monospaced-link")
    return [node], []


def setup(app) -> None:
    app.add_role("mslink", monospaced_link)
    app.add_css_file("custom.css")



from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Final, Literal, Protocol, get_args

import pandas as pd

if TYPE_CHECKING:
    from pyranges import PyRanges, RangeFrame


def return_pyranges_if_possible(
    method: Callable,
) -> Callable:
    """Return a PyRanges object if possible."""

    def wrapper(*args, **kwargs) -> "PyRanges | pd.DataFrame | pd.Series":
        # Call the original groupby method
        result = method(*args, **kwargs)

        if isinstance(result, pd.DataFrame) and set(GENOME_LOC_COLS).issubset(result.columns):
            import pyranges as pr

            return pr.PyRanges(result)

        return result

    return wrapper


# Define the Literal type
VALID_OVERLAP_TYPE = Literal["first", "all", "last", "contained"]

# Extract the options from the Literal type
VALID_OVERLAP_OPTIONS = list(get_args(VALID_OVERLAP_TYPE))
OVERLAP_FIRST, OVERLAP_ALL, OVERLAP_LAST, OVERLAP_CONTAINED = VALID_OVERLAP_OPTIONS

BY_ENTRY_IN_KWARGS = "__by__"

PRESERVE_INDEX_COLUMN = "__old_index__"


CHROM_COL: Final = "Chromosome"
START_COL = "Start"
END_COL = "End"
STRAND_COL = "Strand"
RANGE_COLS = [START_COL, END_COL]
CHROM_AND_STRAND_COLS = [CHROM_COL, STRAND_COL]
GENOME_LOC_COLS = [CHROM_COL, *RANGE_COLS]
GENOME_LOC_COLS_WITH_STRAND = [*GENOME_LOC_COLS, STRAND_COL]

BIGWIG_SCORE_COL = "Score"
FRAME_COL = "Frame"

FORWARD_STRAND: Final = "+"
REVERSE_STRAND: Final = "-"
VALID_GENOMIC_STRAND_TYPE = Literal["+", "-"]
VALID_GENOMIC_STRAND_INFO = [FORWARD_STRAND, REVERSE_STRAND]

VALID_BY_OPTIONS = str | Iterable[str] | None

USE_STRAND_AUTO: Final = "auto"
USE_STRAND_DEFAULT: Final = USE_STRAND_AUTO
VALID_USE_STRAND_TYPE = Literal["auto"] | bool
VALID_USE_STRAND_OPTIONS = [USE_STRAND_AUTO, True, False]

STRAND_BEHAVIOR_AUTO: Final = "auto"
STRAND_BEHAVIOR_SAME: Final = "same"
STRAND_BEHAVIOR_OPPOSITE: Final = "opposite"
STRAND_BEHAVIOR_IGNORE: Final = "ignore"
STRAND_BEHAVIOR_DEFAULT: Final = "auto"
VALID_STRAND_BEHAVIOR_TYPE = Literal["auto", "same", "opposite", "ignore"]
STRICT_STRAND_BEHAVIOR_TYPE = Literal["same", "opposite", "ignore"]
VALID_STRAND_BEHAVIOR_OPTIONS = [
    STRAND_BEHAVIOR_SAME,
    STRAND_BEHAVIOR_OPPOSITE,
    STRAND_BEHAVIOR_AUTO,
    STRAND_BEHAVIOR_IGNORE,
]


JOIN_OUTER: Final = "outer"
JOIN_INNER: Final = "inner"
JOIN_RIGHT: Final = "right"
JOIN_LEFT: Final = "left"
VALID_JOIN_TYPE = Literal["inner", "left", "outer", "right"]
VALID_JOIN_OPTIONS = [JOIN_INNER, JOIN_LEFT, JOIN_OUTER, JOIN_RIGHT]

JOIN_SUFFIX = "_b"
VALID_COMBINE_OPTIONS = Literal["intersect", "union", "swap"]

NEAREST_ANY_DIRECTION: Final = "any"
NEAREST_UPSTREAM: Final = "upstream"
NEAREST_DOWNSTREAM: Final = "downstream"
VALID_NEAREST_TYPE = Literal["any", "upstream", "downstream"]
VALID_NEAREST_OPTIONS = [
    NEAREST_ANY_DIRECTION,
    NEAREST_UPSTREAM,
    NEAREST_UPSTREAM,
    None,
]

VALID_DIRECTION_TYPE = Literal["any", "forward", "backward"]
ANY_DIRECTION = Literal["any"]
FORWARD_DIRECTION = Literal["forward"]
BACKWARD_DIRECTION = Literal["backward"]

TEMP_INDEX_COL = "__temp_index__"
TEMP_TYPE_COL = "__temp_type__"
TEMP_START_COL = "__temp_start__"
TEMP_END_COL = "__temp_end__"
TEMP_STRAND_COL = "__temp_strand__"
TEMP_NAME_COL = "__temp_name__"
TEMP_CUMSUM_COL = "__temp_cumsum__"
TEMP_NUM_COL = "__temp_num__"
TEMP_ID_COL = "__temp_id__"
TEMP_MIN_COL = "__temp_min__"
TEMP_MAX_COL = "__temp_max__"
TEMP_CLUSTER_COL = "__temp_cluster__"
TEMP_LENGTH_COL = "__temp_length__"
TEMP_START_SLACK_COL = "__temp_start_slack__"
TEMP_END_SLACK_COL = "__temp_end_slack__"

DEFAULT_CLUSTER_ID_COL = "Cluster"

VALID_BY_TYPES = str | Iterable[str] | None

PANDAS_COMPRESSION_TYPE = Literal["infer", "gzip", "bz2", "zip", "xz", "zstd"] | dict[str, Any] | None

SKIP_IF_DF_EMPTY_TYPE = Literal["left", "right", "any", "both", False]
SKIP_IF_EMPTY_LEFT: Final = "left"
SKIP_IF_EMPTY_RIGHT: Final = "right"
SKIP_IF_EMPTY_ANY: Final = "any"
SKIP_IF_EMPTY_BOTH: Final = "both"
SKIP_IF_DF_EMPTY_OPTIONS = [False, SKIP_IF_EMPTY_ANY, SKIP_IF_EMPTY_BOTH, SKIP_IF_EMPTY_LEFT, SKIP_IF_EMPTY_RIGHT]
SKIP_IF_DF_EMPTY_DEFAULT = "any"


class UnaryOperation[T: "RangeFrame"](Protocol):
    """A protocol for unary operations on RangeFrames."""

    def __call__(self, df: T, **kwargs: Any) -> "pd.DataFrame":
        """Perform the operation on the RangeFrame.

        Examples: cluster, merge, split, etc.
        """
        ...


class BinaryOperation[T: "RangeFrame"](Protocol):
    """A protocol for binary operations on RangeFrames."""

    def __call__(self, df: T, *, df2: T, **kwargs: Any) -> "pd.DataFrame":
        """Perform the operation on the pair of RangeFrames.

        Examples: overlap, nearest, join, etc.
        """
        ...


class CombineIntervalColumnsOperation(Protocol):
    """A protocol for functions passed to combine_interval_columns.

    A protocol indicating that the operation combines interval columns from a Pyranges object,
    expecting four pd.Series as inputs (starts, ends, starts2, ends2), and returns a tuple of pd.Series (new_starts, new_ends).
    """

    def __call__(
        self,
        starts: pd.Series,
        ends: pd.Series,
        starts2: pd.Series,
        ends2: pd.Series,
    ) -> tuple[pd.Series, pd.Series]:
        """Perform the operation on starts, starts2, ends, ends2 columns and return new start and end columns."""
        ...



from typing import TYPE_CHECKING

import pandas as pd

from pyranges.core.names import CHROM_COL, END_COL, START_COL
from pyranges.core.pyranges_helpers import factorize, mypy_ensure_rangeframe
from pyranges.range_frame.range_frame import RangeFrame

if TYPE_CHECKING:
    import pyranges as pr


def _bounds[T: ("pr.PyRanges", "pd.DataFrame")](df: T, by: list[str]) -> pd.DataFrame:
    if df.empty:
        return df

    col_order = [*dict.fromkeys([col for col in df if col in [*by, START_COL, END_COL]]).keys()]

    import ruranges

    group_ids = factorize(df, by=by)

    idxs, starts, ends, _counts = ruranges.boundary(  # type: ignore[attr-defined]
        groups=group_ids,
        starts=df[START_COL].to_numpy(),
        ends=df[END_COL].to_numpy(),
    )

    ids = df.take(idxs).loc[:, by]  # type: ignore[arg-type]

    result = RangeFrame({START_COL: starts, END_COL: ends} | {_by: ids[_by] for _by in by})[col_order]

    return mypy_ensure_rangeframe(result.reset_index(drop=True))


def _outside_bounds(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    df = df.copy()

    _chromsizes = kwargs.get("chromsizes")

    if isinstance(_chromsizes, pd.DataFrame):
        size_df = _chromsizes.df
        if not size_df.Chromosome.is_unique:
            msg = "Chromosomes must be unique in chromsizes."
            raise ValueError(msg)
        chromsizes = dict(*zip(size_df[CHROM_COL], size_df[END_COL], strict=True))
    else:
        if not isinstance(_chromsizes, dict):
            msg = "chromsizes must be a dictionary or a DataFrame"
            raise TypeError(msg)
        chromsizes = _chromsizes

    size = int(chromsizes[df.Chromosome.iloc[0]])
    clip = kwargs.get("clip", False)
    only_right = kwargs.get("only_right", False)

    ends_outright = df.End > size
    starts_outleft = df.Start < 0

    if not clip:  # i.e. remove
        df = df[~ends_outright] if only_right else df[~ends_outright & ~starts_outleft]

    else:
        starts_outright = df.Start >= size

        if only_right:
            df.loc[ends_outright, "End"] = size

            # removing intervals completely out of bounds
            df = df[~starts_outright]

        else:
            ends_outleft = df.End <= 0

            df.loc[ends_outright, "End"] = size
            df.loc[starts_outleft, "Start"] = 0

            # removing intervals completely out of bounds:
            df = df[~starts_outright & ~ends_outleft]

    return df



import typing
from collections.abc import Iterable

import pandas as pd
from pandas import Series

from pyranges.core import names
from pyranges.core.pyranges_helpers import mypy_ensure_pyranges

if typing.TYPE_CHECKING:
    from pyranges.core.pyranges_main import PyRanges


def empty_df(
    columns: Iterable[str] | None = None,
    dtype: Series | None = None,
    *,
    with_strand: bool = False,
) -> pd.DataFrame:
    """Create an empty DataFrame that is valid as a PyRanges.

    Parameters
    ----------
    columns : Iterable of str, default None
        Columns to create. If None, the default columns Chromosome, Start, and End are used.

    dtype: Series, default None
        Dtype for the columns.

    with_strand: bool, default False
        Whether to create a PyRanges with strand information.

    """
    empty = pd.DataFrame(
        columns=list(columns)
        if columns is not None
        else (names.GENOME_LOC_COLS_WITH_STRAND if with_strand else names.GENOME_LOC_COLS),
    )
    return empty.astype(dtype) if dtype is not None else empty


def empty(
    columns: Iterable[str] | None = None,
    dtype: Series | None = None,
    *,
    strand: bool = False,
) -> "PyRanges":
    """Create an empty PyRanges.

    Parameters
    ----------
    columns : Iterable of str, default None
        Columns to create. If None, the default columns Chromosome, Start, and End are used.

    dtype: Series, default None
        Dtype for the columns.

    strand: bool, default False
        Whether to create a PyRanges with strand information.

    """
    return mypy_ensure_pyranges(empty_df(with_strand=strand, columns=columns, dtype=dtype))



#!/usr/bin/env python3

import argparse
import logging
import shlex
import subprocess
import sys

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


parser = argparse.ArgumentParser(description="This script runs various checks and tests.")
parser.add_argument("-v", "--verbose", action="store_true", help="Show output from programs run.")
parser.add_argument("-l", "--list", action="store_true", help="List all checks and tests and exits.")
args = parser.parse_args()
verbose: bool = args.verbose
show_checks: bool = args.list


def run_command(command, *, show_output: bool) -> int:
    """Run a shell command and return its exit status."""
    result = subprocess.run(command, capture_output=not show_output, text=True, check=False)  # noqa: S603
    return result.returncode


def main() -> int:
    """Run all continuous integration checks."""
    commands = {
        "ruff format": ["ruff", "format", "--check", "--diff", "pyranges"],
        "ruff check": ["ruff", "check", "pyranges"],
        "pyright": ["pyright", "pyranges"],
        "pytest": ["pytest", "--doctest-modules", "pyranges/", "tests/unit/"],
        # "more doctest": ["python", "tests/run_doctest_tutorial_howto.py"],
    }
    if show_checks:
        LOGGER.info("Available checks and tests:")
        for task, command in commands.items():
            LOGGER.info("    %s: %s", task, shlex.join(command))
        sys.exit(0)

    failed_tasks = []

    for task, command in commands.items():
        LOGGER.debug("Running: %s", shlex.join(command))
        retcode = run_command(command, show_output=verbose)
        if retcode != 0:
            LOGGER.debug("Task %s failed!", task)
            failed_tasks.append(task)

    if failed_tasks:
        LOGGER.warning("The following tasks failed:")
        for task in failed_tasks:
            LOGGER.warning("%s:\n    %s", task, shlex.join(commands[task]))
        return 1

    LOGGER.debug("All tasks completed successfully.")
    return 0


if __name__ == "__main__":
    sys.exit(main())



from pyranges.ext.stats import (
    fdr,  # noqa: F401
    fisher_exact,  # noqa: F401
    forbes,  # noqa: F401
    jaccard,  # noqa: F401
    mcc,  # noqa: F401
    relative_distance,  # noqa: F401
    rowbased_pearson,  # noqa: F401
    rowbased_rankdata,  # noqa: F401
    rowbased_spearman,  # noqa: F401
    simes,  # noqa: F401
)



def pytest_configure(config) -> None:
    import pyranges as pr

    pr.options.set_option("max_rows_to_show", 8)
    pr.options.set_option("max_column_names_to_show", 3)
    pr.options.set_option("console_width", 120)



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import ruranges  # Rust extension

from pyranges.core.names import (
    CHROM_COL,
    END_COL,
    START_COL,
    STRAND_COL,
    USE_STRAND_DEFAULT,
    VALID_USE_STRAND_TYPE,
)
from pyranges.core.pyranges_helpers import factorize_binary, mypy_ensure_pyranges

if TYPE_CHECKING:
    from pyranges import PyRanges


def map_to_global(
    local_gr: "PyRanges",
    global_gr: "PyRanges",
    global_on: str,
    *,
    local_on: str = "Chromosome",
    use_strand: VALID_USE_STRAND_TYPE = USE_STRAND_DEFAULT,
) -> "PyRanges":
    """Lift intervals from local_gr to genomic coordinates using global_gr exon annotations.

    Lift intervals in *local_gr* (coordinates relative to a transcript) onto absolute genomic coordinates using the exon annotation in *global_gr*.
    """
    cumsum_start = "_local_start"
    cumsum_end = "_local_end"

    global_cum = global_gr.group_cumsum(
        match_by=global_on,
        use_strand=use_strand,
        cumsum_start_column=cumsum_start,
        cumsum_end_column=cumsum_end,
    )
    ex_df = global_cum.copy()

    ex_df["__global_start__"] = ex_df[START_COL]
    ex_df["__global_end__"] = ex_df[END_COL]
    ex_df[START_COL] = ex_df[cumsum_start]
    ex_df[END_COL] = ex_df[cumsum_end]

    tmp_tx_id = "__txid__"
    local_df = local_gr.copy()

    ex_df[tmp_tx_id] = ex_df[global_on]
    local_df[tmp_tx_id] = local_df[local_on]

    q_tx_code, ex_tx_code = factorize_binary(local_df, ex_df, by=tmp_tx_id)

    ex_chr_code, chr_table = pd.factorize(ex_df[CHROM_COL], sort=False)
    ex_chr_code = ex_chr_code.astype(ex_tx_code.dtype, copy=False)

    ex_local_start = ex_df[START_COL].to_numpy(copy=False)
    ex_local_end = ex_df[END_COL].to_numpy(copy=False)
    ex_genome_start = ex_df["__global_start__"].to_numpy(copy=False)
    ex_genome_end = ex_df["__global_end__"].to_numpy(copy=False)
    ex_fwd = (ex_df[STRAND_COL] == "+").to_numpy(np.bool_, copy=False)

    q_start = local_df[START_COL].to_numpy(copy=False)
    q_end = local_df[END_COL].to_numpy(copy=False)
    q_fwd = (local_df[STRAND_COL] == "+").to_numpy(np.bool_, copy=False)

    keep_idx, out_start, out_end, out_strand_bool = ruranges.map_to_global(
        groups=ex_tx_code,
        starts=ex_local_start,
        ends=ex_local_end,
        strand=ex_fwd,
        groups2=q_tx_code,
        starts2=q_start,
        ends2=q_end,
        chr_code2=ex_chr_code,
        genome_start2=ex_genome_start,
        genome_end2=ex_genome_end,
        strand2=q_fwd,
    )

    if len(keep_idx) == 0:
        return mypy_ensure_pyranges(local_df.head(0))

    max_code = int(ex_tx_code.max()) + 1
    tx2chr_code = np.empty(max_code, dtype=ex_chr_code.dtype)
    tx2chr_code[ex_tx_code] = ex_chr_code
    out_chr_code = tx2chr_code[q_tx_code[keep_idx]]
    out_chr_str = chr_table[out_chr_code]

    mapped = local_gr.take(keep_idx)  # type: ignore[arg-type]
    mapped_df = mapped.copy()

    mapped_df[CHROM_COL] = out_chr_str
    mapped_df[START_COL] = out_start
    mapped_df[END_COL] = out_end
    mapped_df[STRAND_COL] = np.where(out_strand_bool, "+", "-")

    if tmp_tx_id in mapped_df.columns:
        mapped_df = mapped_df.drop(columns=[tmp_tx_id])
        if mapped_df is None:
            raise ValueError

    return mypy_ensure_pyranges(mapped_df)



"""Module of example data.

See Also
--------
pyranges.random : generate random PyRanges

Examples
--------
>>> pr.example_data
Available example data:
-----------------------
example_data.chipseq            : Example ChIP-seq data.
example_data.chipseq_background : Example ChIP-seq data.
example_data.chromsizes         : Example chromsizes data (hg19).
example_data.ensembl_gtf        : Example gtf file from Ensembl.
example_data.f1                 : Example bed file.
example_data.f2                 : Example bed file.
example_data.aorta              : Example ChIP-seq data.
example_data.aorta2             : Example ChIP-seq data.
example_data.ncbi_gff           : Example NCBI GFF data.
example_data.ncbi_fasta         : Example NCBI fasta.
example_data.files              : Return a dict of basenames to file paths of available data.

>>> pr.example_data.f1
  index  |    Chromosome      Start      End  Name         Score  Strand
  int64  |    category        int64    int64  object       int64  category
-------  ---  ------------  -------  -------  ---------  -------  ----------
      0  |    chr1                3        6  interval1        0  +
      1  |    chr1                5        7  interval2        0  -
      2  |    chr1                8        9  interval3        0  +
PyRanges with 3 rows, 6 columns, and 1 index columns.
Contains 1 chromosomes and 2 strands.


"""

import logging
import sys
import tempfile
import typing
from importlib.resources import files
from pathlib import Path
from typing import Any, ClassVar

import pyranges as pr

if typing.TYPE_CHECKING:
    import pyfaidx

    from pyranges.core.pyranges_main import PyRanges

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


# Define a custom class property decorator
class ClassProperty:
    def __init__(self, fget: Any) -> None:
        self.fget = fget
        # Propagate the docstring from the getter function
        self.__doc__ = fget.__doc__

    def __get__(self, instance: Any | None, owner: type[Any]) -> Any:
        return self.fget(owner)


class ExampleData:
    _files: ClassVar[dict[str, Path]] = {}

    def __repr__(self) -> str:
        methods_info = "Available example data:\n-----------------------\n"
        items = [
            (name, prop)
            for name, prop in self.__class__.__dict__.items()
            if isinstance(prop, property) and not name.startswith("_")
        ]
        max_name_len = max(len(name) for name, _ in (*items, ("files", "")))
        for name, prop in items:
            doc_line = prop.fget.__doc__.split("\n")[0] if prop.fget.__doc__ else ""
            methods_info += f"example_data.{name:<{max_name_len}} : {doc_line}\n"
        for name in ["files"]:
            doc_line = self.__class__.__dict__[name].__doc__.strip().split("\n")[0]
            methods_info += f"example_data.{name:<{max_name_len}} : {doc_line}\n"
        return methods_info.strip()

    @ClassProperty
    def files(self) -> dict[str, Path]:
        """Return a dict of basenames to file paths of available data.

        Examples
        --------
        >>> bam = ExampleData.files["smaller.bam"]
        >>> bam.exists()
        True

        """
        if ExampleData._files:
            return ExampleData._files

        paths = []
        for f in files("pyranges").joinpath("data").iterdir():
            if not isinstance(f, Path):
                msg = f"Expected Path, got {type(f)}"
                raise TypeError(msg)
            if "__" not in f.name:
                paths.append(f)
        ExampleData._files = {f.name: Path(f) for f in paths}
        return ExampleData._files

        paths = []
        for f in files("pyranges").joinpath("data").iterdir():
            if not isinstance(f, Path):
                msg = f"Expected Path, got {type(f)}"
                raise TypeError(msg)
            if "__" not in f.name:
                paths.append(f)
        ExampleData._files = {f.name: Path(f) for f in paths}
        return ExampleData._files

    @staticmethod
    def _read_bed_from_string(contents: str) -> "PyRanges":
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pr.read_bed(Path(f.name))

    @staticmethod
    def _read_gtf_from_string(contents: str) -> "PyRanges":
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pr.read_gtf(Path(f.name))

    @property
    def chipseq(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the SICER software.
        """
        contents = """chr8	28510032	28510057	U0	0	-
chr7	107153363	107153388	U0	0	-
chr5	135821802	135821827	U0	0	-
chr14	19418999	19419024	U0	0	-
chr12	106679761	106679786	U0	0	-
chr21	40099618	40099643	U0	0	+
chr8	22714402	22714427	U0	0	-
chr19	19571102	19571127	U0	0	+
chr3	140986358	140986383	U0	0	-
chr10	35419784	35419809	U0	0	-
chr4	98488749	98488774	U0	0	+
chr11	22225193	22225218	U0	0	+
chr1	38457520	38457545	U0	0	+
chr1	80668132	80668157	U0	0	-
chr2	152562484	152562509	U0	0	-
chr4	153155301	153155326	U0	0	+
chr9	120803448	120803473	U0	0	+
chr6	89296757	89296782	U0	0	-
chr1	194245558	194245583	U0	0	+
chr8	57916061	57916086	U0	0	+"""
        return self._read_bed_from_string(contents)

    @property
    def chipseq_background(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the SICER software.
        """
        contents = """chr7	20246668	20246693	U0	0	+
chr1	39036822	39036847	U0	0	+
chr19	47109000	47109025	U0	0	-
chr10	90059861	90059886	U0	0	-
chr3	55648137	55648162	U0	0	+
chr7	91135110	91135135	U0	0	+
chr13	100938475	100938500	U0	0	+
chr3	115816130	115816155	U0	0	+
chr19	43528773	43528798	U0	0	+
chr10	73781101	73781126	U0	0	+"""
        return self._read_bed_from_string(contents)

    @property
    def chromsizes(self) -> "pr.PyRanges":
        """Example chromsizes data (hg19)."""
        contents = """chr1	0	249250621
chr2	0	243199373
chr3	0	198022430
chr4	0	191154276
chr5	0	180915260
chr6	0	171115067
chr7	0	159138663
chrX	0	155270560
chr8	0	146364022
chr9	0	141213431
chr10	0	135534747
chr11	0	135006516
chr12	0	133851895
chr13	0	115169878
chr14	0	107349540
chr15	0	102531392
chr16	0	90354753
chr17	0	81195210
chr18	0	78077248
chr20	0	63025520
chrY	0	59373566
chr19	0	59128983
chr22	0	51304566
chr21	0	48129895
chrM	0	16571"""
        return self._read_bed_from_string(contents)

    @property
    def ensembl_gtf(self) -> "pr.PyRanges":
        """Example gtf file from Ensembl."""
        contents = """#!genome-build GRCh38.p10
#!genome-version GRCh38
#!genome-date 2013-12
#!genome-build-accession NCBI:GCA_000001405.25
#!genebuild-last-updated 2017-06
1	havana	gene	11869	14409	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene";
1	havana	transcript	11869	14409	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; tag "basic"; transcript_support_level "1";
1	havana	exon	11869	12227	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; exon_number "1"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; exon_id "ENSE00002234944"; exon_version "1"; tag "basic"; transcript_support_level "1";
1	havana	exon	12613	12721	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; exon_number "2"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; exon_id "ENSE00003582793"; exon_version "1"; tag "basic"; transcript_support_level "1";
1	havana	exon	13221	14409	.	+	.	gene_id "ENSG00000223972"; gene_version "5"; transcript_id "ENST00000456328"; transcript_version "2"; exon_number "3"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; exon_id "ENSE00002312635"; exon_version "1"; tag "basic"; transcript_support_level "1";
1	havana	exon	112700	112804	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000471248"; transcript_version "1"; exon_number "2"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-203"; transcript_source "havana"; transcript_biotype "lincRNA"; exon_id "ENSE00001957285"; exon_version "1"; transcript_support_level "5";
1	havana	exon	110953	111357	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000471248"; transcript_version "1"; exon_number "3"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-203"; transcript_source "havana"; transcript_biotype "lincRNA"; exon_id "ENSE00001879696"; exon_version "1"; transcript_support_level "5";
1	ensembl	transcript	120725	133723	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; tag "basic"; transcript_support_level "5";
1	ensembl	exon	133374	133723	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; exon_number "1"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; exon_id "ENSE00003748456"; exon_version "1"; tag "basic"; transcript_support_level "5";
1	ensembl	exon	129055	129223	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; exon_number "2"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; exon_id "ENSE00003734824"; exon_version "1"; tag "basic"; transcript_support_level "5";
1	ensembl	exon	120874	120932	.	-	.	gene_id "ENSG00000238009"; gene_version "6"; transcript_id "ENST00000610542"; transcript_version "1"; exon_number "3"; gene_name "AL627309.1"; gene_source "ensembl_havana"; gene_biotype "lincRNA"; transcript_name "AL627309.1-205"; transcript_source "ensembl"; transcript_biotype "lincRNA"; exon_id "ENSE00003740919"; exon_version "1"; tag "basic"; transcript_support_level "5";"""
        return self._read_gtf_from_string(contents)

    @property
    def f1(self) -> "pr.PyRanges":
        """Example bed file."""
        contents = """chr1	3	6	interval1	0	+
chr1	5	7	interval2	0	-
chr1	8	9	interval3	0	+"""
        return self._read_bed_from_string(contents)

    @property
    def f2(self) -> "pr.PyRanges":
        """Example bed file."""
        contents = """chr1	1	2	a	0	+
chr1	6	7	b	0	-"""
        with tempfile.NamedTemporaryFile("w", encoding="utf-8") as f:
            f.write(contents)
            f.flush()
            return pr.read_bed(Path(f.name))

    @property
    def aorta(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the epigenomics roadmap.
        """
        return pr.read_bed(ExampleData.files["aorta.bed"])  # type: ignore[index, misc]

    @property
    def aorta2(self) -> "pr.PyRanges":
        """Example ChIP-seq data.

        From the epigenomics roadmap.
        """
        return pr.read_bed(ExampleData.files["aorta2.bed"])  # type: ignore[index, misc]

    @property
    def ncbi_gff(self) -> "pr.PyRanges":
        """Example NCBI GFF data.

        Subset of the NCBI annotation of D.gyrociliatus assembly GCA_904063045.1.
        """
        return pr.read_gff3(ExampleData.files["ncbi.gff.gz"])  # type: ignore[index, misc]

    @property
    def ncbi_fasta(self) -> "pyfaidx.Fasta":
        """Example NCBI fasta.

        Subset of the NCBI D.gyrociliatus assembly GCA_904063045.1.

        A pyfaidx.Fasta object is returned. If not installed, an exception is raised.
        To retrieve the location of the fasta file, use pyranges.example_data.files['ncbi.fasta']
        """
        try:
            import pyfaidx  # type: ignore[import]
        except ImportError:
            LOGGER.exception(
                "To use this method, pyfaidx must be installed. To get just the fasta file path, use pyranges.example_data.files['ncbi.fasta']. Use `conda install -c bioconda pyfaidx` or `pip install pyfaidx` to install pyfaidx.",
            )
            sys.exit(1)

        # note: example data include ncbi.fasta.fai, the pyfaidx index
        return pyfaidx.Fasta(ExampleData.files["ncbi.fasta"])  # type: ignore[index, misc]


example_data = ExampleData()



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING

import numpy as np

from pyranges.core.names import END_COL, JOIN_SUFFIX, START_COL

if TYPE_CHECKING:
    from pyranges.range_frame.range_frame import RangeFrame

# ------------------------------------------------------------------
# Built-in metric identifiers (all single-column outputs)
# ------------------------------------------------------------------
VALID_METRICS: set[str] = {
    "overlap_length",  # raw overlap in bp
    "fraction",  # overlap / chosen denominator
    "jaccard",  # overlap / union length
    "distance",  # unsigned gap distance
    "overlap",  # boolean flag
    "signed_distance",  # negative = upstream, positive = downstream
    "midpoint_distance",  # absolute distance between midpoints
    "symmetric_coverage",  # 2 * overlap / (len1 + len2)
    "relative_direction",  # same / opposite / unknown (strand aware)
}


def compute_interval_metrics(
    df: RangeFrame,
    metrics: str | Iterable[str] | Mapping[str, str] = "fraction",
    *,
    start: str = START_COL,
    end: str = END_COL,
    start2: str = START_COL + JOIN_SUFFIX,
    end2: str = END_COL + JOIN_SUFFIX,
    denom: str = "first",
) -> RangeFrame:
    """Attach per-row interval-relationship metrics as new columns.

    Parameters
    ----------
    df : RangeFrame
        Frame in which both intervals reside (e.g. the result of a join).
    metrics : str | Iterable[str] | Mapping[str, str], default "fraction"
        Metrics to compute - a name, a list of names, or a mapping
        {metric_name -> output_column_name}.  Valid names are in
        VALID_METRICS.
    start, end : str, default START_COL / END_COL
        Column names holding the first interval coordinates.
    start2, end2 : str, default START_COL + "_b" / END_COL + "_b"
        Column names holding the second interval coordinates.
    denom : {"first", "second", "union"}, default "first"
        Denominator used by the *fraction* metric.

    Returns
    -------
    RangeFrame
        Copy of *df* with one extra column per requested metric.

    """
    # --------------------------------------------------------------
    # 0. Resolve requested operations -> output column names
    # --------------------------------------------------------------
    if isinstance(metrics, Mapping):
        op_to_name = dict(metrics)
    elif isinstance(metrics, str):
        op_to_name = {metrics: metrics}
    else:
        op_to_name = {m: m for m in metrics}

    unknown = set(op_to_name) - VALID_METRICS
    if unknown:
        raise ValueError("Unknown metric(s): " + ", ".join(sorted(unknown)))

    # --------------------------------------------------------------
    # 1. Shared NumPy arrays (vectorised)
    # --------------------------------------------------------------
    s1 = df[start].to_numpy()
    e1 = df[end].to_numpy()
    s2 = df[start2].to_numpy()
    e2 = df[end2].to_numpy()

    len1 = e1 - s1
    len2 = e2 - s2
    overlap = np.maximum(0, np.minimum(e1, e2) - np.maximum(s1, s2))
    union_len = len1 + len2 - overlap
    gap_unsigned = np.where(overlap > 0, 0, np.maximum(s1, s2) - np.minimum(e1, e2))
    signed_gap = np.where(overlap > 0, 0, np.where(e1 <= s2, s2 - e1, -(s1 - e2)))
    midpoint_dist = np.abs(((s1 + e1) / 2) - ((s2 + e2) / 2))
    symmetric_cov = 2 * overlap / np.where(len1 + len2 == 0, np.nan, len1 + len2)
    denom_len = {"first": len1, "second": len2, "union": union_len}[denom]

    # --------------------------------------------------------------
    # 2. Helper for strand-aware metric
    # --------------------------------------------------------------
    def _relative_direction() -> np.ndarray:
        if "Strand" not in df.columns or "Strand" + JOIN_SUFFIX not in df.columns:
            msg = "relative_direction requires 'Strand' and 'Strand_b' columns."
            raise KeyError(msg)
        str1 = df["Strand"].to_numpy(dtype="U1")
        str2 = df["Strand" + JOIN_SUFFIX].to_numpy(dtype="U1")
        return np.where(
            (str1 == ".") | (str2 == "."),
            "unknown",
            np.where(str1 == str2, "same", "opposite"),
        )

    # --------------------------------------------------------------
    # 3. Map each metric name to a no-argument callable
    # --------------------------------------------------------------
    metric_funcs = {
        "overlap_length": lambda: overlap,
        "fraction": lambda: overlap / np.where(denom_len == 0, np.nan, denom_len),
        "jaccard": lambda: overlap / np.where(union_len == 0, np.nan, union_len),
        "distance": lambda: gap_unsigned,
        "overlap": lambda: overlap > 0,
        "signed_distance": lambda: signed_gap,
        "midpoint_distance": lambda: midpoint_dist,
        "symmetric_coverage": lambda: symmetric_cov,
        "relative_direction": _relative_direction,
    }

    # --------------------------------------------------------------
    # 4. Assemble output
    # --------------------------------------------------------------
    result = df.copy()
    for op, col_name in op_to_name.items():
        result[col_name] = metric_funcs[op]()

    return result



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from pyranges.core.names import (
    END_COL,
    RANGE_COLS,
    START_COL,
    VALID_BY_TYPES,
    VALID_OVERLAP_TYPE,
)
from pyranges.core.pyranges_helpers import factorize_binary

if TYPE_CHECKING:
    from pyranges import RangeFrame


def _both_idxs(
    df: "RangeFrame",
    df2: "RangeFrame",
    *,
    by: VALID_BY_TYPES,
    multiple: VALID_OVERLAP_TYPE = "all",
    contained: bool = False,
    slack: int = 0,
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    import ruranges

    f1, f2 = factorize_binary(df, df2, by)

    idx1, idx2 = ruranges.overlaps(
        groups=f1,
        starts=df.Start.to_numpy(),
        ends=df.End.to_numpy(),
        groups2=f2,
        starts2=df2.Start.to_numpy(),
        ends2=df2.End.to_numpy(),
        multiple=multiple,
        contained=contained,
        slack=slack,
    )
    return idx1, idx2


def _overlap(
    df: "RangeFrame",
    df2: "RangeFrame",
    *,
    by: VALID_BY_TYPES,
    multiple: VALID_OVERLAP_TYPE = "all",
    contained: bool = False,
    slack: int = 0,
) -> pd.DataFrame:
    idx1, _ = _both_idxs(
        df=df,
        df2=df2,
        by=by,
        multiple=multiple,
        contained=contained,
        slack=slack,
    )
    return df.take(idx1)  # type: ignore[]


def _intersect(
    df: "RangeFrame",
    df2: "RangeFrame",
    *,
    by: list[str],
    multiple: VALID_OVERLAP_TYPE = "all",
    slack: int = 0,
    contained: bool = False,
) -> pd.DataFrame:
    idx1, idx2 = _both_idxs(
        df=df,
        df2=df2,
        by=by,
        multiple=multiple,
        contained=contained,
        slack=slack,
    )

    rf, rf2 = df.take(idx1), df2.take(idx2).loc[:, RANGE_COLS]  # type: ignore[arg-type]

    new_starts = np.where(rf.Start.to_numpy() > rf2.Start.to_numpy(), rf.Start, rf2.Start)

    new_ends = np.where(rf.End.to_numpy() < rf2.End.to_numpy(), rf.End, rf2.End)

    rf.loc[:, START_COL] = new_starts
    rf.loc[:, END_COL] = new_ends

    return rf



# minimal setup.py to be able to use the -e flag (pip install -e .)

from setuptools import find_packages, setup

setup(
    package_data={"pyranges": ["data/*.bam", "data/ncbi.*", "data/bigwig.bw"]},
    include_package_data=True,
    packages=find_packages(),
)



import logging
import sys
import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
import pandas as pd

from pyranges.core.names import (
    CHROM_COL,
    END_COL,
    FORWARD_STRAND,
    FRAME_COL,
    REVERSE_STRAND,
    START_COL,
    STRAND_COL,
    TEMP_CUMSUM_COL,
    TEMP_INDEX_COL,
    TEMP_LENGTH_COL,
)
from pyranges.core.pyranges_helpers import arg_to_list, mypy_ensure_pyranges

if TYPE_CHECKING:
    import pyranges as pr

## Joan Pallares started this function
#  Marco Mariotti later optimized it, made it independent from pyranges

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

DIRECTION_OPTIONS = Literal["up", "down"]

STARTS_NUCLEOTIDE_SEQ = ["ATG"]
STOPS_NUCLEOTIDE_SEQ = ["TAG", "TGA", "TAA"]

N_PRINTED_NON_MULTIPLE_3 = 10


def calculate_frame(p: "pr.PyRanges", transcript_id: str | list[str], frame_col: str = "Frame") -> "pr.PyRanges":
    """Calculate the frame of genomic intervals, assuming all are coding sequences (CDS), and add it as column.

    A stranded
    After this, the input Pyranges will contain an added "Frame" column, which determines the nucleotide of the CDS
    that is the first base of a codon.Resulting values are in range between 0 and 2 included.
    0 indicates that the first nucleotide of that interval is the first base of a codon,
    1 indicates the second base and 2 indicates the third base.
    While the 5'-most interval of each transcript has always 0 frame, the following ones may have any of these values.

    Parameters
    ----------
    p : PyRanges
        Input CDS intervals.

    transcript_id : str or list of str
        Column(s) to group by the intervals: coding exons belonging to the same transcript have the same values in this/these column(s).

    frame_col : str, default 'Frame'
        Name of the column to store the frame values.

    Returns
    -------
    PyRanges

    Examples
    --------
    >>> import pyranges as pr
    >>> p = pr.PyRanges({"Chromosome": [1,1,1,2,2],
    ...                   "Strand": ["+","+","+","-","-"],
    ...                   "Start": [1,31,52,101,201],
    ...                   "End": [10,45,90,130,218],
    ...                   "transcript_id": ["t1","t1","t1","t2","t2"]})
    >>> p
      index  |      Chromosome  Strand      Start      End  transcript_id
      int64  |           int64  object      int64    int64  object
    -------  ---  ------------  --------  -------  -------  ---------------
          0  |               1  +               1       10  t1
          1  |               1  +              31       45  t1
          2  |               1  +              52       90  t1
          3  |               2  -             101      130  t2
          4  |               2  -             201      218  t2
    PyRanges with 5 rows, 5 columns, and 1 index columns.
    Contains 2 chromosomes and 2 strands.

    >>> pr.orfs.calculate_frame(p, transcript_id=['transcript_id'])
      index  |      Chromosome  Strand      Start      End  transcript_id      Frame
      int64  |           int64  object      int64    int64  object             int64
    -------  ---  ------------  --------  -------  -------  ---------------  -------
          0  |               1  +               1       10  t1                     0
          1  |               1  +              31       45  t1                     0
          2  |               1  +              52       90  t1                     2
          3  |               2  -             101      130  t2                     2
          4  |               2  -             201      218  t2                     0
    PyRanges with 5 rows, 6 columns, and 1 index columns.
    Contains 2 chromosomes and 2 strands.

    """
    if not p.strand_valid:
        msg = "Strand must be valid to run calculate_frame."
        raise AssertionError(msg)

    gr = p.copy()

    # Column to save the initial index
    gr[TEMP_INDEX_COL] = np.arange(len(p))

    # Filtering for desired columns
    sorted_p = gr.get_with_loc_columns([TEMP_INDEX_COL, *arg_to_list(transcript_id)])

    # Sorting by 5' (Intervals on + are sorted by ascending order and - are sorted by descending order)
    sorted_p = sorted_p.sort_ranges()

    # Creating a column saving the length for the intervals (for selenoprofiles and ensembl)
    sorted_p[TEMP_LENGTH_COL] = sorted_p.lengths()

    # Creating a column saving the cumulative length for the intervals
    sorted_p[TEMP_CUMSUM_COL] = sorted_p.groupby(transcript_id)[TEMP_LENGTH_COL].cumsum()

    # Creating a frame column
    sorted_p[FRAME_COL] = (sorted_p[TEMP_CUMSUM_COL] - sorted_p[TEMP_LENGTH_COL]) % 3

    # Appending the Frame of sorted_p by the index of p
    sorted_p = mypy_ensure_pyranges(sorted_p.sort_values(by=TEMP_INDEX_COL))

    gr[frame_col] = sorted_p[FRAME_COL]

    return gr.drop_and_return(TEMP_INDEX_COL, axis=1)


def extend_orfs(  # noqa: C901,PLR0912,PLR0915
    p: "pr.PyRanges",
    fasta_path: str,
    transcript_id: str | list[str] | None = None,
    *,
    direction: Iterable[DIRECTION_OPTIONS] | DIRECTION_OPTIONS | None = None,
    starts: list[str] = STARTS_NUCLEOTIDE_SEQ,
    stops: list[str] = STOPS_NUCLEOTIDE_SEQ,
    keep_off_bounds: bool = False,
    record_extensions: bool = False,
    chunk_size: int = 900,
    verbose: bool = False,
) -> "pr.PyRanges":
    r"""Extend PyRanges intervals to form complete open reading frames.

    The input intervals are extended their next Stop codon downstream, and to their leftmost Start codon upstream
    before encountering a Stop.

    Parameters
    ----------
    p : PyRanges
        Input CDS intervals.

    fasta_path : location of the Fasta file from which the sequences
        for the extensions will be retrieved.

    transcript_id : str or list of str or None
        Name(s) of column(s) to group intervals into transcripts

    starts : list containing the nucleotide pattern to look for upstream.
        if not provided, ORFs are delimited by stops
        Default ['ATG']

    stops : list containing the nucleotide pattern to look for downstream.
        Default ['TAG', 'TGA', 'TAA']

    direction : whether the extension should be upstream ('up'), downstream
        ('down') or both. Default (None) means: ['up', 'down']

    keep_off_bounds : if True, those intervals that reached out of bounds during extension
        without finding any stop are returned in their largest (3-nt multiple) extension.
        In this case, these intervals will not begin with a start or end with a stop

    record_extensions : if True, add columns extension_up and extension_down
        with the extensions amounts. Default: False

    chunk_size : the amount of nucleotides to be extended on each iteration. Does not affect output, only speed/memory.
        Default 900.

    verbose : if True, print information about the progress of the extension.
        Default: False

    Note
    ----

    This function requires the library pyfaidx, it can be installed with
    ``conda install -c bioconda pyfaidx`` or ``pip install pyfaidx``.

    Sorting the PyRanges is likely to improve the speed.
    Intervals on the negative strand will be reverse complemented.

    Examples
    --------
    >>> import pyranges as pr
    >>> p = pr.PyRanges({"Chromosome": ['seq1'], "Start":[20], "End":[29], "Strand" : ["+"]})
    >>> p
      index  |    Chromosome      Start      End  Strand
      int64  |    object          int64    int64  object
    -------  ---  ------------  -------  -------  --------
          0  |    seq1               20       29  +
    PyRanges with 1 rows, 4 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    >>> #            *       ^       ^      ... ... ...          *       #  ... = p interval
    >>> seq1 = " AA TAA TGT ATG GTA ATG GGC GCC GGG ATT CCA CAG TAA GTG C".replace(' ', '')
    >>> tmp_handle = open("temp.fasta", "w+")
    >>> _ = tmp_handle.write(">seq1\n")
    >>> _ = tmp_handle.write(seq1+'\n')
    >>> tmp_handle.close()

    >>> p.get_sequence("temp.fasta")
    0    GCCGGGATT
    Name: Sequence, dtype: object

    >>> ep = pr.orfs.extend_orfs(p, fasta_path="temp.fasta")
    >>> ep
      index  |    Chromosome      Start      End  Strand
      int64  |    object          int64    int64  object
    -------  ---  ------------  -------  -------  --------
          0  |    seq1                8       38  +
    PyRanges with 1 rows, 4 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    >>> ep.get_sequence("temp.fasta")
    0    ATGGTAATGGGCGCCGGGATTCCACAGTAA
    Name: Sequence, dtype: object

    >>> pr.orfs.extend_orfs(p, fasta_path="temp.fasta", record_extensions=True)
      index  |    Chromosome      Start      End  Strand      extension_up    extension_down
      int64  |    object          int64    int64  object             int64             int64
    -------  ---  ------------  -------  -------  --------  --------------  ----------------
          0  |    seq1                8       38  +                     12                 9
    PyRanges with 1 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    Extending only in one direction:

    >>> pr.orfs.extend_orfs(p, fasta_path="temp.fasta", direction='up')
      index  |    Chromosome      Start      End  Strand
      int64  |    object          int64    int64  object
    -------  ---  ------------  -------  -------  --------
          0  |    seq1                8       29  +
    PyRanges with 1 rows, 4 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    With starts=[], any codon can be used as a start (i.e. ORFs defined as stop-delimited sequences):

    >>> ep=pr.orfs.extend_orfs(p, fasta_path="temp.fasta", starts=[])
    >>> ep
      index  |    Chromosome      Start      End  Strand
      int64  |    object          int64    int64  object
    -------  ---  ------------  -------  -------  --------
          0  |    seq1                5       38  +
    PyRanges with 1 rows, 4 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    >>> ep.get_sequence("temp.fasta")
    0    TGTATGGTAATGGGCGCCGGGATTCCACAGTAA
    Name: Sequence, dtype: object

    Example with multi-exon input intervals
    Intervals on the negative strand are extended accordingly

    >>> #                   :    *           ^      ... .        ..      *      # ... = p interval
    >>> # reverse complement: C TAG CGT TTG ATG TTG GGC CAG GTG TTT CAG TAG CCC GG
    >>> seq2 = " CC GGG CTA CTG AAA CAC CTG GCC CAA CAT CAA ACG CTA G".replace(' ', '')
    >>> tmp_handle = open("temp1.fasta", "w+")
    >>> _ = tmp_handle.write(">seq2\n")
    >>> _ = tmp_handle.write(seq2+'\n')
    >>> tmp_handle.close()


    >>> np = pr.PyRanges({"Chromosome": ['seq2']*2, "Start":[19, 11], "End":[23, 13],
    ...                   "Strand" : ["-"]*2, "ID":["a", "a"]})
    >>> np
      index  |    Chromosome      Start      End  Strand    ID
      int64  |    object          int64    int64  object    object
    -------  ---  ------------  -------  -------  --------  --------
          0  |    seq2               19       23  -         a
          1  |    seq2               11       13  -         a
    PyRanges with 2 rows, 5 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    >>> np.get_sequence("temp1.fasta")
    0    GGCC
    1      TT
    Name: Sequence, dtype: object

    >>> ep = pr.orfs.extend_orfs(np, fasta_path="temp1.fasta", transcript_id='ID')
    >>> ep.get_sequence("temp1.fasta")
    0    ATGTTGGGCC
    1      TTCAGTAG
    Name: Sequence, dtype: object


    A sequence with no in-frame stops after the input interval before the end:

    >>> #             *       ^       ^      ... ... ...                   #  ... = p interval
    >>> seq1b = " AA TAA TGT ATG GTA ATG GGC GCC GGG ATT CCA CAG AAA GTG C".replace(' ', '')
    >>> tmp_handle = open("temp2.fasta", "w+")
    >>> _ = tmp_handle.write(">seq1\n")
    >>> _ = tmp_handle.write(seq1b+'\n')
    >>> tmp_handle.close()

    >>> pr.orfs.extend_orfs(p, fasta_path="temp2.fasta", record_extensions=True)
      index  |    Chromosome      Start      End  Strand      extension_up    extension_down
      int64  |    object          int64    int64  object             int64             int64
    -------  ---  ------------  -------  -------  --------  --------------  ----------------
          0  |    seq1                8       29  +                     12                 0
    PyRanges with 1 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    Showcasing keep_off_bounds:

    >>> ep=pr.orfs.extend_orfs(p, fasta_path="temp2.fasta", record_extensions=True, keep_off_bounds=True)
    >>> ep
      index  |    Chromosome      Start      End  Strand      extension_up    extension_down
      int64  |    object          int64    int64  object             int64             int64
    -------  ---  ------------  -------  -------  --------  --------------  ----------------
          0  |    seq1                8       41  +                     12                12
    PyRanges with 1 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.
    >>> ep.get_sequence("temp2.fasta")
    0    ATGGTAATGGGCGCCGGGATTCCACAGAAAGTG
    Name: Sequence, dtype: object

    A sequence with no in-frame stops BEFORE the input interval:

    >>> #                     ^       ^      ... ... ...          *        #  ... = p interval
    >>> seq1c = " AA TAC TGT ATG GTA ATG GGC GCC GGG ATT CCA CAG TAA GTG C".replace(' ', '')
    >>> tmp_handle = open("temp3.fasta", "w+")
    >>> _ = tmp_handle.write(">seq1\n")
    >>> _ = tmp_handle.write(seq1c+'\n')
    >>> tmp_handle.close()

    >>> pr.orfs.extend_orfs(p, fasta_path="temp3.fasta", record_extensions=True)
      index  |    Chromosome      Start      End  Strand      extension_up    extension_down
      int64  |    object          int64    int64  object             int64             int64
    -------  ---  ------------  -------  -------  --------  --------------  ----------------
          0  |    seq1                8       38  +                     12                 9
    PyRanges with 1 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    >>> ep = pr.orfs.extend_orfs(p, fasta_path="temp3.fasta", record_extensions=True, keep_off_bounds=True)
    >>> ep
      index  |    Chromosome      Start      End  Strand      extension_up    extension_down
      int64  |    object          int64    int64  object             int64             int64
    -------  ---  ------------  -------  -------  --------  --------------  ----------------
          0  |    seq1                2       38  +                     18                 9
    PyRanges with 1 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes and 1 strands.

    >>> ep.get_sequence("temp3.fasta")
    0    TACTGTATGGTAATGGGCGCCGGGATTCCACAGTAA
    Name: Sequence, dtype: object

    """
    try:
        import pyfaidx  # type: ignore[import]
    except ImportError:
        LOGGER.exception(
            "To use extend_orfs, pyfaidx must be installed. Use `conda install -c bioconda pyfaidx` or `pip install pyfaidx` to install pyfaidx.",
        )
        sys.exit(1)

    def pverbose(msg: Any) -> None:
        if verbose:
            LOGGER.info(msg)

    # Sanity Checks
    if (chunk_size % 3) != 0:
        msg = "Chunk size must be a multiple of three."
        raise AssertionError(msg)

    if not all(len(pattern) == len(STARTS_NUCLEOTIDE_SEQ[0]) for pattern in starts + stops):
        msg = "Ensure that all patterns have a length of 3 nt."
        raise AssertionError(msg)

    if not stops:
        msg = "At least one stop codon must be provided."
        raise AssertionError(msg)

    direction = direction if direction is not None else ["up", "down"]

    p = p.copy()

    if not p.strand_valid:
        msg = "Intervals must be have valid strands to call extend_orfs"
        raise AssertionError(msg)

    if transcript_id is None:
        cds_id = "Custom_ID"
        p[cds_id] = np.arange(len(p))  # Generates a New Column
    else:
        cds_id = transcript_id

    if isinstance(cds_id, str):  # ensure it is a list
        cds_id = [cds_id]

    ##################
    # Show a warning if some transcript lengths are not divisible by 3:
    p["__length"] = p[END_COL] - p[START_COL]
    module_3 = p.groupby(by=cds_id).__length.sum().mod(3)  # noqa: SLF001
    ndv = module_3[module_3 != 0].index
    if len(ndv) > 0:
        warnings.warn("\nWARNING! Some input lenghts are not divisible by 3:", stacklevel=2)
        warnings.warn(
            " ".join(ndv[:N_PRINTED_NON_MULTIPLE_3]) + (" and others " if len(ndv) > N_PRINTED_NON_MULTIPLE_3 else ""),
            stacklevel=2,
        )

    p = p.drop_and_return(["__length"], axis=1)
    ##################

    # Load Sequence Data from a Fasta file
    fs = pyfaidx.Fasta(fasta_path)  # pyfaidx_fasta object, fasta sequences

    # get a minimal interval per group, with min start and max end
    minp = (
        (p[[STRAND_COL, START_COL, END_COL, CHROM_COL, *cds_id]])
        .groupby(cds_id, as_index=False)
        .agg(
            {
                STRAND_COL: "first",
                START_COL: "min",
                END_COL: "max",
                CHROM_COL: "first",
            },
        )
    )
    ncds = len(minp)
    nexons = len(p)

    # add chromosome sizes as column
    minp = minp.merge(
        pd.DataFrame(
            {
                CHROM_COL: [chrom for chrom, _ in fs.items()],
                "__chromsize": [len(fs[chrom]) for chrom, _ in fs.items()],
            },
        ),
        on=CHROM_COL,
    ).set_index(cds_id)

    if len(minp) != ncds:
        msg = "ERROR some sequences were not found in the pyfaidx object"
        raise AssertionError(msg)

    ######################  Extend Sequence Upstream ###########################

    if "up" in direction or direction == "up":
        pup = minp.copy()
        pup["__out_of_bounds"] = False
        pup["__up_stop_dist"] = -1  # -1 means not found
        pup["__up_start_dist"] = -1  # -1 means not found
        selector = pd.Series(
            data=True,
            index=pup.index,
        )  # initialized as all True; keeps track of intervals with work to do
        ic = 0  # iteration counter

        while (selector).any():
            pverbose(
                f"Upstream | iteration {ic},  left to do: {selector.sum()}",
            )
            pverbose(pup[selector])
            # selector= ... everything which is not done nor out of bounds
            _extend_df(pup, "5", chunk_size, selector)
            _subseq_df(pup, "5", chunk_size, selector)
            _correct_bounds(pup, adjust="5")
            _get_seqs(pup, fs, selector)  # seqs length is only selector

            ## looking for a stop if didn't find it already
            z = pup.loc[selector, "__seq"].apply(lambda x: _find_rightmost_stop(x, stops))

            found_stop_selector = selector.copy()  # made to index pup
            found_stop_selector[found_stop_selector] = z != -1
            pup.loc[found_stop_selector, "__up_stop_dist"] = z[z != -1] + ic * chunk_size

            no_stop_off_bounds = selector.copy()
            no_stop_off_bounds[no_stop_off_bounds] = z == -1

            if keep_off_bounds:
                # focus on intervals that reached out of bounds in this iteration and did not find a stop:
                #  record their max extension as up_stop_dist and __up_start_dist so they are considered whether starts was provided or not
                # btw this means these intervals can be identified by looking at rows with __up_start_dist == __up_stop_dist

                pup.loc[no_stop_off_bounds, "__up_stop_dist"] = (
                    pup.loc[no_stop_off_bounds, END_COL] - pup.loc[no_stop_off_bounds, START_COL]
                ) + ic * chunk_size
                pup.loc[no_stop_off_bounds, "__up_start_dist"] = pup.loc[no_stop_off_bounds, "__up_stop_dist"]

            ## looking for a start
            if starts:
                w = pup.loc[selector, ["__seq", "__up_stop_dist"]].apply(
                    lambda x: _find_leftmost_start(x, starts, chunk_size),
                    axis=1,
                )
                ## if start is not found now, but was found in a previous iteration, the previous will be kept (not overwriting a -1)
                found_start_selector = selector.copy()  # made to index pup
                found_start_selector[found_start_selector] = w != -1
                if not keep_off_bounds:
                    pup.loc[found_start_selector, "__up_start_dist"] = w[w != -1] + ic * chunk_size
                else:
                    found_start_and_stop_selector = found_start_selector & ~no_stop_off_bounds
                    pup.loc[found_start_and_stop_selector, "__up_start_dist"] = w[w != -1] + ic * chunk_size
            else:
                # if no starts are provided, the start is whatever comes after the stop
                pup.loc[found_stop_selector, "__up_start_dist"] = pup.loc[found_stop_selector, "__up_stop_dist"] - 3

            pup.loc[found_stop_selector, "__seq"] = ""  ## frees memory
            pverbose(" ---> After iteration:")
            pverbose(pup[selector])

            selector = (pup.__up_stop_dist == -1) & ~(pup.__out_of_bounds)  # noqa: SLF001
            ic += 1

        ext_up = pup["__up_start_dist"].rename("extension_up")

        ext_up[ext_up == -1] = 0
        del pup

    else:  # No extension upstream (i.e. the extension is 0)
        ext_up = pd.Series(0, index=minp.index)

    ######################  Extend Sequence Downstream #########################
    if "down" in direction or direction == "down":
        pdo = minp.copy()
        pdo["__out_of_bounds"] = False
        pdo["__down_stop_dist"] = -1  # -1 means not found
        selector = pd.Series(
            data=True,
            index=pdo.index,
        )  # initialized as all True; keeps track of intervals with work to do
        ic = 0  # iteration counter

        while (selector).any():
            pverbose(
                f"Downstream | iteration {ic},  left to do: {selector.sum()}",
            )
            pverbose(pdo[selector])
            # selector= ... everything which is not done nor out of bounds
            _extend_df(pdo, "3", chunk_size, selector)
            _subseq_df(pdo, "3", chunk_size, selector)
            _correct_bounds(pdo, adjust="3")
            _get_seqs(pdo, fs, selector)  # seqs length is only selector

            ## looking for a stop if didn't find it already
            z = pdo.loc[selector, "__seq"].apply(lambda x: _find_leftmost_stop(x, stops))
            found_stop_selector = selector.copy()  # made to index pdo
            found_stop_selector[found_stop_selector] = z != -1
            pdo.loc[found_stop_selector, "__down_stop_dist"] = z[z != -1] + ic * chunk_size

            if keep_off_bounds:
                # focus on intervals that reached out of bounds in this iteration and did not find a stop:
                #  record their max extension as down_stop_dist

                no_stop_off_bounds = selector.copy()
                no_stop_off_bounds[no_stop_off_bounds] = z == -1

                pdo.loc[no_stop_off_bounds, "__down_stop_dist"] = (
                    pdo.loc[no_stop_off_bounds, END_COL] - pdo.loc[no_stop_off_bounds, START_COL]
                ) + ic * chunk_size

            pdo.loc[found_stop_selector, "__seq"] = ""  ## frees memory
            pverbose(" ---> After iteration:")
            pverbose(pdo[selector])
            selector = (pdo.__down_stop_dist == -1) & ~(pdo.__out_of_bounds)  # noqa: SLF001
            ic += 1

        ext_down = pdo["__down_stop_dist"].rename("extension_down")
        ext_down[ext_down == -1] = 0
        del pdo

    else:  # No extension downstream (i.e. the extension is 0)
        ext_down = pd.Series(data=0, index=minp.index)

    ### Extensions have been determined. Now let's apply them to the original DF
    p["__order"] = np.arange(len(p))
    zp = mypy_ensure_pyranges(
        p.merge(
            pd.DataFrame({"__extension_up": ext_up, "__extension_down": ext_down}),
            left_on=cds_id,
            right_index=True,
        ),
    )
    if len(zp) != nexons:
        msg = "ERROR malformed gene structures"
        raise AssertionError(msg)

    p = zp

    p = p.sort_ranges()
    _extend_groups(p, cds_id)
    p = mypy_ensure_pyranges(p.sort_values("__order"))  # restore order

    # getting ready to return
    if record_extensions:
        p = mypy_ensure_pyranges(
            p.rename(
                columns={
                    "__extension_up": "extension_up",
                    "__extension_down": "extension_down",
                },
            ),
        )
        to_drop = ["__order"]
    else:
        to_drop = ["__extension_up", "__extension_down", "__order"]

    if transcript_id is None:
        to_drop.extend(cds_id)

    return mypy_ensure_pyranges(p.drop_and_return(to_drop, axis=1))


# MM # efficients methods to work inplace. All methods with df as input have one row per CDS group
def _extend_df(df, direction, extension, selector) -> None:
    """Inplace."""
    if direction == "5":
        df.loc[selector & (df.Strand == FORWARD_STRAND), [START_COL]] -= extension
        df.loc[selector & (df.Strand == "-"), [END_COL]] += extension
    elif direction == "3":
        df.loc[selector & (df.Strand == REVERSE_STRAND), [START_COL]] -= extension
        df.loc[selector & (df.Strand == FORWARD_STRAND), [END_COL]] += extension


def _extend_groups(dfg, cds_id) -> None:
    """In place, dfg may contain multiple exons."""
    first_exon_indices = dfg.groupby(cds_id).apply(lambda x: x.index[0]).reset_index(drop=True)
    last_exon_indices = dfg.groupby(cds_id).apply(lambda x: x.index[-1]).reset_index(drop=True)

    # boolean Series to select first exons
    first_exon_selector = dfg.index.isin(set(first_exon_indices))
    last_exon_selector = dfg.index.isin(set(last_exon_indices))

    # boolean Series to select positive and negative strand
    sp = dfg.Strand == FORWARD_STRAND
    sm = dfg.Strand == REVERSE_STRAND

    # Find the intersection to get, e.g. first exon on positive strand
    fe_sp = first_exon_selector & sp
    fe_sm = first_exon_selector & sm
    le_sp = last_exon_selector & sp
    le_sm = last_exon_selector & sm

    dfg.loc[fe_sp, START_COL] -= dfg.loc[fe_sp, "__extension_up"]
    dfg.loc[le_sp, END_COL] += dfg.loc[le_sp, "__extension_down"]
    dfg.loc[fe_sm, END_COL] += dfg.loc[fe_sm, "__extension_up"]
    dfg.loc[le_sm, START_COL] -= dfg.loc[le_sm, "__extension_down"]


def _get_seqs(df, fs, selector) -> None:
    """Creates/updates column __seq."""
    seqs = []
    d = df[selector]
    for start, end, strand, chrom in zip(d.Start, d.End, d.Strand, d.Chromosome, strict=False):
        _fasta = fs[chrom]
        if strand == REVERSE_STRAND:
            seqs.append((-_fasta[start:end]).seq.upper())  # reverse complement
        elif strand == FORWARD_STRAND:
            seqs.append(_fasta[start:end].seq.upper())
    df.loc[selector, "__seq"] = seqs


def _correct_bounds(df, adjust=None) -> None:
    """Similar to pyranges genome_bounds, inplace, but see below.

    but also: adjust the 5' or 3' end of intervals so they are multiples of 3
    and set boolean attribute __out_of_bounds.
    """
    out_end = df[END_COL] > df["__chromsize"]
    df.loc[out_end, END_COL] = df.loc[out_end, [END_COL, "__chromsize"]].min(axis=1)
    out_start = df[START_COL] < 0
    df.loc[out_start, START_COL] = 0

    sp = df.Strand == FORWARD_STRAND
    sm = df.Strand == REVERSE_STRAND

    if adjust == "5":
        df.loc[out_start & sp, START_COL] += (df.loc[out_start & sp, END_COL] - df.loc[out_start & sp, START_COL]).mod(
            3,
        )
        df.loc[out_end & sm, END_COL] -= (df.loc[out_end & sm, END_COL] - df.loc[out_end & sm, START_COL]).mod(3)

    elif adjust == "3":
        df.loc[out_end & sp, END_COL] -= (df.loc[out_end & sp, END_COL] - df.loc[out_end & sp, START_COL]).mod(3)
        df.loc[out_start & sm, START_COL] += (df.loc[out_start & sm, END_COL] - df.loc[out_start & sm, START_COL]).mod(
            3,
        )

    df.loc[(out_end | out_start), "__out_of_bounds"] = True


def _subseq_df(df, from_end, amount, selector) -> None:
    """Similar to pyranges subseq but inplace."""
    sp = selector & (df.Strand == FORWARD_STRAND)
    sm = selector & (df.Strand == REVERSE_STRAND)
    if from_end == "5":
        df.loc[sp, END_COL] = df.loc[sp, START_COL] + amount
        df.loc[sm, START_COL] = df.loc[sm, END_COL] - amount
    elif from_end == "3":
        df.loc[sp, START_COL] = df.loc[sp, END_COL] - amount
        df.loc[sm, END_COL] = df.loc[sm, START_COL] + amount


def _find_rightmost_stop(seq, motifs) -> int:
    """Return the distance from end of seq of the 1st codon position of the rightmost stop; e.g. 3 is for last codon."""
    for i in range(len(seq) - 3, -1, -3):
        if seq[i : i + 3] in motifs:
            return len(seq) - i
    return -1


def _find_leftmost_start(row, motifs, chunk_size) -> int:
    """Return distance from end of seq of 1st codon position of leftmost start AFTER the pos in column __up_stop_dist."""
    seq = row["__seq"]
    stop_pos = row["__up_stop_dist"]
    start_search = (3 + len(seq) - (stop_pos % chunk_size)) if stop_pos != -1 else 0
    # we search starting from the position next to the stop
    for i in range(start_search, len(seq), 3):
        if seq[i : i + 3] in motifs:
            return len(seq) - i
    return -1


def _find_leftmost_stop(seq, motifs) -> int:
    """Return the position (distance from the start of the string) of the 3rd codon position of the leftmost stop.

    e.g. 3 is for the first codon.
    """
    for i in range(0, len(seq), 3):
        if seq[i : i + 3] in motifs:
            return i + 3
    return -1



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from pyranges.core.names import CHROM_COL, END_COL, START_COL
from pyranges.core.pyranges_helpers import mypy_ensure_rangeframe

if TYPE_CHECKING:
    from pyranges.range_frame.range_frame import RangeFrame


def _complement(
    df: "RangeFrame",
    *,
    by: list[str],
    slack: int = 0,
    chromsizes_col: str | None = None,
    chromsizes: "dict[str | int, int] | None" = None,
    include_first_interval: bool = False,
) -> "RangeFrame":
    import ruranges

    from pyranges.range_frame.range_frame import RangeFrame

    if df.empty:
        return df

    col_order = [col for col in df if col in [*by, START_COL, END_COL]]

    factorized = pd.Series(np.zeros(len(df), dtype=np.uint32)) if not by else df.groupby(by).ngroup().astype(np.uint32)

    if chromsizes and chromsizes_col:
        chrom_lens = df[chromsizes_col].replace(chromsizes)
        chrom_lens = pd.concat([factorized, chrom_lens], axis=1).drop_duplicates()
        chrom_len_ids = chrom_lens[0].to_numpy()
        chrom_lens = chrom_lens[chromsizes_col].to_numpy()
    else:
        chrom_len_ids = np.array([], dtype=np.uint32)
        chrom_lens = np.array([], dtype=np.int64)

    chrs, start, end, idxs = ruranges.complement(
        groups=factorized.to_numpy(),
        starts=df.Start.to_numpy(),
        ends=df.End.to_numpy(),
        slack=slack,
        chrom_len_ids=chrom_len_ids,  # type: ignore[arg-type]
        chrom_lens=chrom_lens,
        include_first_interval=include_first_interval,
    )

    ids = df.take(idxs)  # type: ignore[arg-type]

    result = RangeFrame({CHROM_COL: chrs, START_COL: start, END_COL: end} | {_by: ids[_by] for _by in by})[col_order]

    return mypy_ensure_rangeframe(result.reset_index(drop=True))
